/**
 * Minimal EventEmitter interface that is molded against the Node.js
 * EventEmitter interface.
 */
export default class EventEmitter {
    private _events;
    /**
     * Return a list of assigned event listeners.
     */
    listeners(event: string): Function[];
    /**
     * Emit an event to all registered event listeners.
     */
    emit(event: string, ...args: any[]): boolean;
    /**
     * Callback executor for overriding.
     */
    protected callListener(listener: Function, context: Object, params: any[]): any;
    /**
     * Remove event listeners.
     */
    removeListener(event: string, fn: Function, once?: boolean): this;
    /**
     * Register a new EventListener for the given event.
     */
    on(event: string, fn: Function, context?: Object): this;
    /**
     * Add an EventListener that's only called once.
     */
    once(event: string, fn: Function, context?: Object): this;
    /**
     * Remove all listeners or only the listeners for the specified event.
     */
    removeAllListeners(event: string): this;
}
