import AsyncMachine from './asyncmachine';
import { MutationTypes, IQueueRow, TransitionStepTypes, IStateStruct, ITransitionStep, IBind, IEmit } from './types';
/**
 * The Transition class is responsible for encapsulating a single mutation
 * for a single machine. In can be created by a different machine than it's
 * mutating. End users usually don't have to deal with it at all, as the most
 * important data it carries for them is exposed as `instance.to()` and
 * `instance.from()` methods and it's events are also emitted on the
 * machine itself.
 *
 * TODO freeze the attributes
 */
export default class Transition {
    id: any;
    source_machine: AsyncMachine<any, IBind, IEmit>;
    private events;
    before: string[];
    states: string[];
    enters: string[];
    exits: string[];
    accepted: boolean;
    row: IQueueRow;
    steps: ITransitionStep[];
    cancelled: boolean;
    readonly machine: AsyncMachine<any, IBind, IEmit>;
    readonly auto: boolean;
    readonly type: MutationTypes;
    readonly requested_states: string[];
    readonly params: any[];
    constructor(source_machine: AsyncMachine<any, IBind, IEmit>, row: IQueueRow);
    exec(): boolean;
    setupAccepted(): void;
    resolveRelations(states: string[]): void;
    prepareAutoStates(): IQueueRow | null;
    protected parseAddRelation(states: string[]): string[];
    parseRequires_(states: string[]): string[];
    /**
     * Returns the subset of states which block the state name.
     */
    isStateBlocked_(states: string[], name: string): string[];
    orderStates_(states: string[]): void;
    removeDuplicateStates_(states: string[]): string[];
    setupExitEnter(): void;
    self(): boolean;
    enter(to: string): any;
    exit(from: string): boolean;
    transitionExec_(from: string, to: string | null, method: string, params?: string[]): any;
    processPostTransition(): void;
    processPostTransitionException(state: string, is_enter: boolean): void;
    /**
     * Marks a steps relation between two states during the transition.
     */
    addStep(target: string | IStateStruct, source?: string | IStateStruct | null, type?: TransitionStepTypes, data?: any): void;
    /**
     * Marks a steps relation between two states during the transition.
     */
    addStepData(target: string | IStateStruct, source?: string | IStateStruct | null, type?: TransitionStepTypes, data?: any): ITransitionStep;
    /**
     * Same as [[addStep]], but produces a step for many targets.
     */
    addStepsFor(targets: string[] | IStateStruct[], source?: string | IStateStruct | null, type?: TransitionStepTypes, data?: any): void;
    /**
     * Produces a readable list of steps states.
     *
     * Example:
     * ```
     * A   REQUESTED
     * D   REQUESTED
     * A   SET
     * D   SET
     * D -> B   RELATION   add
     * D -> C   RELATION   add
     * B   SET
     * C   SET
     * E -> D   RELATION   drop
     * E   DROP
     * ```
     *
     * TODO loose casts once condition guards work again
     */
    toString(): string;
}
