"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var MutationTypes;
(function (MutationTypes) {
    MutationTypes[MutationTypes["DROP"] = 0] = "DROP";
    MutationTypes[MutationTypes["ADD"] = 1] = "ADD";
    MutationTypes[MutationTypes["SET"] = 2] = "SET";
})(MutationTypes = exports.MutationTypes || (exports.MutationTypes = {}));
/**
 * Queue enum defining param positions in queue's entries.
 */
var QueueRowFields;
(function (QueueRowFields) {
    QueueRowFields[QueueRowFields["STATE_CHANGE_TYPE"] = 0] = "STATE_CHANGE_TYPE";
    QueueRowFields[QueueRowFields["STATES"] = 1] = "STATES";
    QueueRowFields[QueueRowFields["PARAMS"] = 2] = "PARAMS";
    QueueRowFields[QueueRowFields["AUTO"] = 3] = "AUTO";
    QueueRowFields[QueueRowFields["TARGET"] = 4] = "TARGET";
})(QueueRowFields = exports.QueueRowFields || (exports.QueueRowFields = {}));
class Deferred {
    constructor() {
        this.promise = new Promise((resolve, reject) => {
            this.resolve = resolve;
            this.reject = reject;
        });
    }
}
exports.Deferred = Deferred;
var StateRelations;
(function (StateRelations) {
    StateRelations["AFTER"] = "after";
    StateRelations["ADD"] = "add";
    StateRelations["REQUIRE"] = "require";
    StateRelations["DROP"] = "drop";
})(StateRelations = exports.StateRelations || (exports.StateRelations = {}));
var TransitionStepTypes;
(function (TransitionStepTypes) {
    TransitionStepTypes[TransitionStepTypes["RELATION"] = 1] = "RELATION";
    TransitionStepTypes[TransitionStepTypes["TRANSITION"] = 4] = "TRANSITION";
    TransitionStepTypes[TransitionStepTypes["SET"] = 8] = "SET";
    TransitionStepTypes[TransitionStepTypes["DROP"] = 16] = "DROP";
    TransitionStepTypes[TransitionStepTypes["NO_SET"] = 32] = "NO_SET";
    TransitionStepTypes[TransitionStepTypes["REQUESTED"] = 64] = "REQUESTED";
    TransitionStepTypes[TransitionStepTypes["CANCEL"] = 128] = "CANCEL";
    TransitionStepTypes[TransitionStepTypes["PIPE"] = 256] = "PIPE";
})(TransitionStepTypes = exports.TransitionStepTypes || (exports.TransitionStepTypes = {}));
var StateStructFields;
(function (StateStructFields) {
    StateStructFields[StateStructFields["MACHINE_ID"] = 0] = "MACHINE_ID";
    StateStructFields[StateStructFields["STATE_NAME"] = 1] = "STATE_NAME";
})(StateStructFields = exports.StateStructFields || (exports.StateStructFields = {}));
var TransitionStepFields;
(function (TransitionStepFields) {
    TransitionStepFields[TransitionStepFields["STATE"] = 0] = "STATE";
    TransitionStepFields[TransitionStepFields["SOURCE_STATE"] = 1] = "SOURCE_STATE";
    TransitionStepFields[TransitionStepFields["TYPE"] = 2] = "TYPE";
    TransitionStepFields[TransitionStepFields["DATA"] = 3] = "DATA";
})(TransitionStepFields = exports.TransitionStepFields || (exports.TransitionStepFields = {}));
/**
 * By default piped are "_state" and "_end" methods, not the negotiation ones.
 * Use the PipeFlags.NEGOTIATION flag to pipe "_enter" and "_exit" methods, and
 * thus, to participate in the state negotiation. This mode DOES NOT guarantee,
 * that the state was successfully negotiated in the source machine.
 *
 * To invert the state, use the PipeFlags.INVERT flag.
 *
 * To append the transition to the local queue (instead of the target
 * machine's one), use the PipeFlags.LOCAL_QUEUE. This will alter the
 * transition order.
 */
var PipeFlags;
(function (PipeFlags) {
    PipeFlags[PipeFlags["NEGOTIATION"] = 1] = "NEGOTIATION";
    PipeFlags[PipeFlags["INVERT"] = 4] = "INVERT";
    PipeFlags[PipeFlags["LOCAL_QUEUE"] = 8] = "LOCAL_QUEUE";
    // TODO write tests for those
    PipeFlags[PipeFlags["FINAL"] = 16] = "FINAL";
    PipeFlags[PipeFlags["NEGOTIATION_ENTER"] = 32] = "NEGOTIATION_ENTER";
    PipeFlags[PipeFlags["NEGOTIATION_EXIT"] = 64] = "NEGOTIATION_EXIT";
    PipeFlags[PipeFlags["FINAL_ENTER"] = 128] = "FINAL_ENTER";
    PipeFlags[PipeFlags["FINAL_EXIT"] = 256] = "FINAL_EXIT";
})(PipeFlags = exports.PipeFlags || (exports.PipeFlags = {}));
exports.PipeFlagsLabels = {
    NEGOTIATION: 'neg',
    INVERT: 'inv',
    LOCAL_QUEUE: 'loc',
    FINAL: 'fin',
    NEGOTIATION_ENTER: 'neg_enter',
    NEGOTIATION_EXIT: 'neg_exit',
    FINAL_ENTER: 'fin_enter',
    FINAL_EXIT: 'fin_exit'
};
class TransitionException extends Error {
    constructor(err, transition) {
        super();
        this.err = err;
        this.transition = transition;
    }
}
exports.TransitionException = TransitionException;
class NonExistingStateError extends Error {
    constructor(name) {
        super('NonExistingStateError: ' + name);
    }
}
exports.NonExistingStateError = NonExistingStateError;
//# sourceMappingURL=types.js.map